"""
A collection of input and output EOTasks
"""

from .geometry_io import GeopediaVectorImportTask, VectorImportTask
from .geopedia import AddGeopediaFeatureTask
from .raster_io import ExportToTiffTask, ImportFromTiffTask
from .sentinelhub_process import (
    SentinelHubDemTask,
    SentinelHubEvalscriptTask,
    SentinelHubInputTask,
    SentinelHubSen2corTask,
    get_available_timestamps,
)

__version__ = "1.4.0"



"""
Module containing tasks used for reading and writing to disk

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2018-2019 William Ouellette (TomTom)
Copyright (c) 2019 Drew Bollinger (DevelopmentSeed)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import datetime as dt
import functools
import logging
import warnings
from abc import ABCMeta
from typing import Any, BinaryIO, List, Optional, Tuple, Union

import fs
import numpy as np
import rasterio
import rasterio.warp
from affine import Affine
from fs.base import FS
from fs.osfs import OSFS
from fs.tempfs import TempFS
from fs_s3fs import S3FS
from rasterio.io import DatasetReader
from rasterio.session import AWSSession
from rasterio.windows import Window, from_bounds

from sentinelhub import CRS, BBox, SHConfig, parse_time_interval

from eolearn.core import EOPatch, FeatureType
from eolearn.core.core_tasks import IOTask
from eolearn.core.exceptions import EORuntimeWarning
from eolearn.core.utils.fs import get_base_filesystem_and_path, get_full_path

LOGGER = logging.getLogger(__name__)


class BaseRasterIoTask(IOTask, metaclass=ABCMeta):  # noqa: B024
    """Base abstract class for raster IO tasks"""

    def __init__(
        self,
        feature,
        folder: str,
        *,
        filesystem: Optional[FS] = None,
        image_dtype: Optional[Union[np.dtype, type]] = None,
        no_data_value: Optional[float] = None,
        create: bool = False,
        config: Optional[SHConfig] = None,
    ):
        """
        :param feature: Feature which will be exported or imported
        :param folder: A path to a main folder containing all image, potentially in its subfolders. If `filesystem`
            parameter is defined, then `folder` should be a path relative to filesystem object. Otherwise, it should be
            an absolute path.
        :param filesystem: An existing filesystem object. If not given it will be initialized according to `folder`
            parameter.
        :param image_dtype: A data type of data in exported images or data imported from images.
        :param no_data_value: When exporting this is the NoData value of pixels in exported images.
            When importing this value is assigned to the pixels with NoData.
        :param create: If the filesystem path doesn't exist this flag indicates to either create it or raise an error.
        :param config: A configuration object with AWS credentials. By default, is set to None and in this case the
            default configuration will be taken.
        """
        self.feature = self.parse_feature(feature)
        self.image_dtype = image_dtype
        self.no_data_value = no_data_value

        if filesystem is None:
            filesystem, folder = get_base_filesystem_and_path(folder, create=create, config=config)

        super().__init__(folder, filesystem=filesystem, create=create, config=config)

    def _get_filename_paths(self, filename_template: Union[str, List[str]], timestamps: List[dt.datetime]) -> List[str]:
        """From a filename "template" and base path on the filesystem it generates full paths to tiff files. The paths
        are still relative to the filesystem object.
        """
        if isinstance(filename_template, str):
            filename_path = fs.path.join(self.filesystem_path, filename_template)
            filename_paths = self._generate_paths(filename_path, timestamps)

        elif isinstance(filename_template, list):
            filename_paths = []
            for timestamp_index, path in enumerate(filename_template):
                filename_path = fs.path.join(self.filesystem_path, path)
                if len(filename_template) == len(timestamps):
                    filename_paths.extend(self._generate_paths(filename_path, [timestamps[timestamp_index]]))
                elif not timestamps:
                    filename_paths.extend(self._generate_paths(filename_path, timestamps))
                else:
                    raise ValueError(
                        "The number of provided timestamps does not match the number of provided filenames."
                    )
        else:
            raise TypeError(
                f"The 'filename' parameter must either be a list or a string, but {filename_template} found"
            )

        if self._create_path:
            paths_to_create = {fs.path.dirname(filename_path) for filename_path in filename_paths}
            for filename_path in paths_to_create:
                self.filesystem.makedirs(filename_path, recreate=True)

        return filename_paths

    @classmethod
    def _generate_paths(cls, path_template: str, timestamps: List[dt.datetime]) -> List[str]:
        """Uses a filename path template to create a list of actual filename paths."""
        if not cls._has_tiff_file_extension(path_template):
            path_template = f"{path_template}.tif"

        if not timestamps:
            return [path_template]

        if "*" in path_template:
            path_template = path_template.replace("*", "%Y%m%dT%H%M%S")

        if timestamps[0].strftime(path_template) == path_template:
            return [path_template]

        return [timestamp.strftime(path_template) for timestamp in timestamps]

    @staticmethod
    def _has_tiff_file_extension(path: str) -> bool:
        """Checks if path ends with a tiff file extension."""
        path = path.lower()
        return path.endswith(".tif") or path.endswith(".tiff")


class ExportToTiffTask(BaseRasterIoTask):
    """Task exports specified feature to GeoTIFF.

    The task can export also features with sizes of both time and band dimension greater than `1`. When exporting
    only multiple times or only multiple bands, the GeoTIFF channels are in the expected order. However, when
    exporting multiple times and bands, the order obeys the following pattern:

    T(1)B(1), T(1)B(2), ..., T(1)B(N), T(2)B(1), T(2)B(2), ..., T(2)B(N), ..., ..., T(M)B(N)

    where T and B are the time and band indices of the array, and M and N are the lengths of these indices,
    respectively.
    """

    def __init__(
        self,
        feature,
        folder: str,
        *,
        date_indices: Union[List[int], Tuple[int, int], Tuple[dt.datetime, dt.datetime], Tuple[str, str], None] = None,
        band_indices: Union[List[int], Tuple[int, int], None] = None,
        crs: Union[CRS, int, str, None] = None,
        fail_on_missing: bool = True,
        compress: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        :param feature: A feature to be exported.
        :param folder: A path to a main folder containing all image, potentially in its subfolders. If `filesystem`
            parameter is defined, then `folder` should be a path relative to filesystem object. Otherwise, it should be
            an absolute path.
        :param date_indices: Indices of those time frames from the give feature that will be exported to a tiff image.
            It can be either a list of indices or a tuple of `2` indices defining an interval of indices or a tuple of
            `2` datetime object also defining a time interval. By default, all time frames will be exported.
        :param band_indices: Indices of those bands from the given feature that will be exported to a tiff image. It
            can be either a list of indices or a tuple of `2` indices defining an interval of indices. By default, all
            bands will be exported.
        :param crs: A CRS in which to reproject the feature before writing it to GeoTIFF. By default, no reprojection
            will be done.
        :param fail_on_missing: A flag to specify if the task should fail if a feature is missing or if it should
            just show a warning.
        :param compress: A type of compression that rasterio should apply to an exported image.
        :param kwargs: Keyword arguments to be propagated to `BaseRasterIoTask`.
        """
        super().__init__(feature, folder=folder, create=True, **kwargs)

        self.date_indices = date_indices
        self.band_indices = band_indices
        self.crs = None if crs is None else CRS(crs)
        self.fail_on_missing = fail_on_missing
        self.compress = compress

    def _prepare_image_array(
        self, data_array: np.ndarray, timestamps: List[dt.datetime], feature: Tuple[FeatureType, str]
    ) -> np.ndarray:
        """Collects a feature from EOPatch and prepares the array of an image which will be rasterized. The resulting
        array has shape (channels, height, width) and is of correct dtype.
        """
        data_array = self._reduce_by_bands(data_array)

        feature_type, _ = feature
        if feature_type.is_temporal():
            data_array = self._reduce_by_time(data_array, timestamps)
        else:
            # add temporal dimension
            data_array = np.expand_dims(data_array, axis=0)

        if not feature_type.is_spatial():
            # add height and width dimensions
            data_array = np.expand_dims(np.expand_dims(data_array, axis=1), axis=1)

        data_array = self._set_export_dtype(data_array, feature)

        time_dim, height, width, band_dim = data_array.shape
        new_shape = (time_dim * band_dim, height, width)
        data_array = np.moveaxis(data_array, -1, 1).reshape(new_shape)

        return data_array

    def _reduce_by_bands(self, array: np.ndarray) -> np.ndarray:
        """Reduces the array by selecting a subset of bands."""
        if self.band_indices is None:
            return array

        if isinstance(self.band_indices, list):
            if [band for band in self.band_indices if not isinstance(band, int)]:
                raise ValueError(f"Invalid format in {self.band_indices} list, expected integers")
            return array[..., self.band_indices]

        if isinstance(self.band_indices, tuple):
            if tuple(map(type, self.band_indices)) != (int, int):
                raise ValueError(f"Invalid format in {self.band_indices} tuple, expected integers")
            return array[..., self.band_indices[0] : self.band_indices[1] + 1]

        raise ValueError(f"Invalid format in {self.band_indices}, expected tuple or list")

    def _reduce_by_time(self, array: np.ndarray, timestamps: List[dt.datetime]) -> np.ndarray:
        """Reduce array by selecting a subset of times."""
        if self.date_indices is None:
            return array

        if isinstance(self.date_indices, list):
            if [date for date in self.date_indices if not isinstance(date, int)]:
                raise ValueError(f"Invalid format in {self.date_indices} list, expected integers")
            return array[np.array(self.date_indices), ...]

        if isinstance(self.date_indices, tuple):
            dates = np.array(timestamps)
            if tuple(map(type, self.date_indices)) == (int, int):
                start_date = dates[self.date_indices[0]]
                end_date = dates[self.date_indices[1]]
            elif tuple(map(type, self.date_indices)) == (str, str):
                start_date, end_date = parse_time_interval(self.date_indices)
            elif tuple(map(type, self.date_indices)) == (dt.datetime, dt.datetime):
                start_date = self.date_indices[0]
                end_date = self.date_indices[1]
            else:
                raise ValueError(f"Invalid format in {self.date_indices} tuple, expected ints, strings, or datetimes")
            return array[np.nonzero(np.where((dates >= start_date) & (dates <= end_date), dates, 0))[0]]

        raise ValueError(f"Invalid format in {self.date_indices}, expected tuple or list")

    def _set_export_dtype(self, data_array: np.ndarray, feature: Tuple[FeatureType, str]) -> np.ndarray:
        """To a given array it sets a dtype in which data will be exported"""
        image_dtype = data_array.dtype if self.image_dtype is None else self.image_dtype

        if image_dtype == np.int64:
            image_dtype = np.int32
            warnings.warn(
                (
                    f"Data from feature {feature} cannot be exported to tiff with dtype numpy.int64. Will export "
                    "as numpy.int32 instead"
                ),
                EORuntimeWarning,
            )

        if image_dtype == data_array.dtype:
            return data_array
        return data_array.astype(image_dtype)

    def _get_source_and_destination_params(
        self, data_array: np.ndarray, bbox: BBox
    ) -> Tuple[Tuple[str, Affine], Tuple[str, Affine], Tuple[int, int]]:
        """Calculates source and destination CRS and transforms. Additionally, it returns destination height and width
        """
        _, height, width = data_array.shape

        src_crs = bbox.crs.ogc_string()
        src_transform = rasterio.transform.from_bounds(*bbox, width=width, height=height)

        if self.crs:
            dst_crs = self.crs.ogc_string()
            dst_transform, dst_width, dst_height = rasterio.warp.calculate_default_transform(
                src_crs, dst_crs, width, height, *bbox
            )
        else:
            dst_crs = src_crs
            dst_transform = src_transform
            dst_width, dst_height = width, height

        return (src_crs, src_transform), (dst_crs, dst_transform), (dst_height, dst_width)

    def _export_tiff(
        self,
        image_array: np.ndarray,
        filesystem: FS,
        path: str,
        channel_count: int,
        dst_crs: str,
        dst_transform: Affine,
        dst_height: int,
        dst_width: int,
        src_crs: str,
        src_transform: Affine,
    ) -> None:
        """Export an EOPatch feature to tiff based on input channel range."""
        with rasterio.Env(), filesystem.openbin(path, "w") as file_handle:  # noqa: SIM117
            with rasterio.open(
                file_handle,
                "w",
                driver="GTiff",
                width=dst_width,
                height=dst_height,
                count=channel_count,
                dtype=image_array.dtype,
                nodata=self.no_data_value,
                transform=dst_transform,
                crs=dst_crs,
                compress=self.compress,
            ) as dst:
                if dst_crs == src_crs:
                    dst.write(image_array)
                else:
                    for idx in range(channel_count):
                        rasterio.warp.reproject(
                            source=image_array[idx, ...],
                            destination=rasterio.band(dst, idx + 1),
                            src_transform=src_transform,
                            src_crs=src_crs,
                            dst_transform=dst_transform,
                            dst_crs=dst_crs,
                            resampling=rasterio.warp.Resampling.nearest,
                        )

    def execute(self, eopatch: EOPatch, *, filename: Union[str, List[str], None] = "") -> EOPatch:
        """Execute method

        :param eopatch: An input EOPatch
        :param filename: A filename with a path to a tiff file. The path has to be a relative to the filesystem object
            and path from `folder` initialization parameter. If the file extension of the image file is not provided,
            it will default to `.tif`. If a "*" wildcard or a datetime.strftime substring (e.g. "%Y%m%dT%H%M%S") is
            provided in the path, an EOPatch feature will be split over multiple GeoTIFFs where each one will correspond
            to a single timestamp. Alternatively, a list of paths can be provided, one for each timestamp. Set this
            parameter to `None` to disable exporting.
        :return: Unchanged input EOPatch
        """
        if filename is None:
            return eopatch

        if self.feature not in eopatch:
            error_msg = f"Feature {self.feature[1]} of type {self.feature[0]} was not found in EOPatch"
            LOGGER.warning(error_msg)
            if self.fail_on_missing:
                raise ValueError(error_msg)
            return eopatch
        if eopatch.bbox is None:
            raise ValueError(
                "Given EOPatch is missing a bounding box and therefore no feature can be exported to GeoTIFF"
            )

        image_array = self._prepare_image_array(eopatch[self.feature], eopatch.timestamp, self.feature)

        (
            (src_crs, src_transform),
            (dst_crs, dst_transform),
            (dst_height, dst_width),
        ) = self._get_source_and_destination_params(image_array, eopatch.bbox)

        filename_paths = self._get_filename_paths(filename, eopatch.timestamp)

        with self.filesystem as filesystem:
            export_function = functools.partial(
                self._export_tiff,
                filesystem=filesystem,
                dst_crs=dst_crs,
                dst_transform=dst_transform,
                dst_height=dst_height,
                dst_width=dst_width,
                src_crs=src_crs,
                src_transform=src_transform,
            )

            channel_count = image_array.shape[0]
            if len(filename_paths) > 1:
                single_channel_count = channel_count // len(eopatch.timestamp)
                for timestamp_index, path in enumerate(filename_paths):
                    time_slice_array = image_array[
                        timestamp_index * single_channel_count : (timestamp_index + 1) * single_channel_count, ...
                    ]
                    export_function(image_array=time_slice_array, path=path, channel_count=single_channel_count)
            else:
                export_function(image_array=image_array, path=filename_paths[0], channel_count=channel_count)

        return eopatch


class ImportFromTiffTask(BaseRasterIoTask):
    """Task for importing data from a GeoTIFF file into an EOPatch

    The task can take an existing EOPatch and read the part of GeoTIFF image, which intersects with its bounding
    box, into a new feature. But if no EOPatch is given it will create a new EOPatch, read entire GeoTIFF image into a
    feature and set a bounding box of the new EOPatch.

    Note that if GeoTIFF file is not completely spatially aligned with location of given EOPatch it will try to fit it
    as good as possible. However, it will not do any spatial resampling or interpolation on GeoTIFF data.
    """

    def __init__(
        self, feature, folder: str, *, use_vsi: bool = False, timestamp_size: Optional[int] = None, **kwargs: Any
    ):
        """
        :param feature: EOPatch feature into which data will be imported
        :param folder: A directory containing image files or a path of an image file
        :param use_vsi: A flag to define if reading should be done with GDAL/rasterio virtual system (VSI)
            functionality. The flag only has an effect when the task is used to read an image from a remote storage
            (i.e. AWS S3 bucket). For a performance improvement it is recommended to set this to `True` when reading a
            smaller chunk of a larger image, especially if it is a Cloud-optimized GeoTIFF (COG). In other cases the
            reading might be faster if the flag remains set to `False`.
        :param timestamp_size: In case data will be imported into a time-dependant feature this parameter can be used to
            specify time dimension. If not specified, time dimension will be the same as size of `FeatureType.TIMESTAMP`
            feature. If `FeatureType.TIMESTAMP` does not exist it will be set to 1.
            When converting data into a feature channels of given tiff image should be in order
            T(1)B(1), T(1)B(2), ..., T(1)B(N), T(2)B(1), T(2)B(2), ..., T(2)B(N), ..., ..., T(M)B(N)
            where T and B are the time and band indices.
        :param kwargs: Keyword arguments to be propagated to `BaseRasterIoTask`.
        """
        super().__init__(feature, folder=folder, **kwargs)

        self.use_vsi = use_vsi
        self.timestamp_size = timestamp_size

    def _get_session(self, filesystem: FS) -> AWSSession:
        """Creates a session object with credentials from a config object."""
        if not isinstance(filesystem, S3FS):
            raise NotImplementedError("A rasterio session for VSI reading for now only works for AWS S3 filesystems")

        return AWSSession(
            aws_access_key_id=filesystem.aws_access_key_id,
            aws_secret_access_key=filesystem.aws_secret_access_key,
            aws_session_token=filesystem.aws_session_token,
            region_name=filesystem.region,
            endpoint_url=filesystem.endpoint_url,
        )

    def _load_from_image(self, path: str, filesystem: FS, bbox: Optional[BBox]) -> Tuple[np.ndarray, Optional[BBox]]:
        """The method decides in what way data will be loaded the image.

        The method always uses `rasterio.Env` to suppress any low-level warnings. In case of a local filesystem
        benchmarks show that without `filesystem.openbin` in some cases `rasterio` can read much faster. Otherwise,
        reading depends on `use_vsi` flag. In some cases where a sub-image window is read and the image is in certain
        format (e.g. COG), benchmarks show that reading with virtual system (VSI) is much faster. In other cases,
        reading with `filesystem.openbin` is faster.
        """
        if isinstance(filesystem, (OSFS, TempFS)):
            full_path = filesystem.getsyspath(path)
            with rasterio.Env():
                return self._read_image(full_path, bbox)

        if self.use_vsi:
            session = self._get_session(filesystem)
            with rasterio.Env(session=session):
                full_path = get_full_path(filesystem, path)
                return self._read_image(full_path, bbox)

        with rasterio.Env(), filesystem.openbin(path, "r") as file_handle:
            return self._read_image(file_handle, bbox)

    def _read_image(self, file_object: Union[str, BinaryIO], bbox: Optional[BBox]) -> Tuple[np.ndarray, Optional[BBox]]:
        """Reads data from the image."""
        with rasterio.open(file_object) as src:
            src: DatasetReader

            read_window, read_bbox = self._get_reading_window_and_bbox(src, bbox)
            boundless_reading = read_window is not None
            return src.read(window=read_window, boundless=boundless_reading, fill_value=self.no_data_value), read_bbox

    @staticmethod
    def _get_reading_window_and_bbox(
        reader: DatasetReader, bbox: Optional[BBox]
    ) -> Tuple[Optional[Window], Optional[BBox]]:
        """Provides a reading window for which data will be read from image. If it returns `None` this means that the
        whole image should be read. Those cases are when bbox is not defined, image is not geo-referenced, or
        bbox coordinates exactly match image coordinates. Additionally, it provides a bounding box of reading window
        if an image is geo-referenced."""
        image_crs = reader.crs
        image_transform = reader.transform
        image_bounds = reader.bounds
        if image_crs is None or image_transform is None or image_bounds is None:
            return None, bbox

        image_bbox = BBox(list(image_bounds), crs=image_crs.to_epsg())
        if bbox is None:
            return None, image_bbox

        original_bbox = bbox
        if bbox.crs is not image_bbox.crs:
            bbox = bbox.transform(image_crs.to_epsg())

        if bbox == image_bbox:
            return None, original_bbox

        return from_bounds(*iter(bbox), transform=image_transform), original_bbox

    def _load_data(self, filename_paths: List[str], initial_bbox: Optional[BBox]) -> Tuple[np.ndarray, Optional[BBox]]:
        """Load data from images, join them, and provide their bounding box."""
        data_per_path: List[np.ndarray] = []
        final_bbox: Optional[BBox] = None

        with self.filesystem as filesystem:
            for path in filename_paths:
                data, bbox = self._load_from_image(path, filesystem, initial_bbox)
                data_per_path.append(data)

                if bbox is None:
                    continue
                if final_bbox and bbox != final_bbox:
                    raise RuntimeError(
                        "Given images have different geo-references. They can't be imported into an EOPatch that"
                        " doesn't have a bounding box."
                    )
                final_bbox = bbox

        return np.concatenate(data_per_path, axis=0), final_bbox

    def execute(self, eopatch: Optional[EOPatch] = None, *, filename: Optional[str] = "") -> EOPatch:
        """Execute method which adds a new feature to the EOPatch

        :param eopatch: input EOPatch or None if a new EOPatch should be created
        :type eopatch: EOPatch or None
        :param filename: filename of tiff file or None if entire path has already been specified in `folder` parameter
            of task initialization.
        :type filename: str, list of str or None
        :return: New EOPatch with added raster layer
        :rtype: EOPatch
        """
        if filename is None:
            if eopatch is None:
                raise ValueError("Both eopatch and filename parameters cannot be set to None")
            return eopatch

        feature_type, feature_name = self.feature
        eopatch = eopatch or EOPatch()

        filename_paths = self._get_filename_paths(filename, eopatch.timestamp)

        data, bbox = self._load_data(filename_paths, eopatch.bbox)

        if eopatch.bbox is None:
            eopatch.bbox = bbox

        if self.image_dtype is not None:
            data = data.astype(self.image_dtype)

        if not feature_type.is_spatial():
            data = data.flatten()

        if feature_type.is_timeless():
            data = np.moveaxis(data, 0, -1)
        else:
            channels = data.shape[0]

            times = self.timestamp_size
            if times is None:
                times = len(eopatch.timestamp) if eopatch.timestamp else 1

            if channels % times != 0:
                raise ValueError(
                    "Cannot import as a time-dependant feature because the number of tiff image channels "
                    "is not divisible by the number of timestamps"
                )

            data = data.reshape((times, channels // times) + data.shape[1:])
            data = np.moveaxis(data, 1, -1)

        eopatch[feature_type][feature_name] = data

        return eopatch



"""
Module for adding vector data from various sources

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Žiga Lukšič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import abc
import logging
from typing import Any, Optional, Union

import boto3
import fiona
import geopandas as gpd
from fiona.session import AWSSession
from fs.base import FS
from fs_s3fs import S3FS

from sentinelhub import CRS, BBox, GeopediaFeatureIterator, SHConfig

from eolearn.core import EOPatch, EOTask, FeatureTypeSet, pickle_fs, unpickle_fs
from eolearn.core.types import FeatureSpec
from eolearn.core.utils.fs import get_base_filesystem_and_path, get_full_path

LOGGER = logging.getLogger(__name__)


class _BaseVectorImportTask(EOTask, metaclass=abc.ABCMeta):
    """Base Vector Import Task, implementing common methods"""

    def __init__(
        self, feature: FeatureSpec, reproject: bool = True, clip: bool = False, config: Optional[SHConfig] = None
    ):
        """
        :param feature: A vector feature into which to import data
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :param config: A configuration object with credentials
        """
        self.feature = self.parse_feature(feature, allowed_feature_types=FeatureTypeSet.VECTOR_TYPES)
        self.config = config or SHConfig()
        self.reproject = reproject
        self.clip = clip

    @abc.abstractmethod
    def _load_vector_data(self, bbox: Optional[BBox]) -> gpd.GeoDataFrame:
        """Loads vector data given a bounding box"""

    def _reproject_and_clip(self, vectors: gpd.GeoDataFrame, bbox: Optional[BBox]) -> gpd.GeoDataFrame:
        """Method to reproject and clip vectors to the EOPatch crs and bbox"""

        if self.reproject:
            if not bbox:
                raise ValueError("To reproject vector data, eopatch.bbox has to be defined!")

            vectors = vectors.to_crs(bbox.crs.pyproj_crs())

        if self.clip:
            if not bbox:
                raise ValueError("To clip vector data, eopatch.bbox has to be defined!")

            bbox_crs = bbox.crs.pyproj_crs()
            if vectors.crs != bbox_crs:
                raise ValueError("To clip, vectors should be in same CRS as EOPatch bbox!")

            extent = gpd.GeoSeries([bbox.geometry], crs=bbox_crs)
            vectors = gpd.clip(vectors, extent, keep_geom_type=True)

        return vectors

    def execute(self, eopatch: Optional[EOPatch] = None, *, bbox: Optional[BBox] = None) -> EOPatch:
        """
        :param eopatch: An existing EOPatch. If none is provided it will create a new one.
        :param bbox: A bounding box for which to load data. By default, if none is provided, it will take a bounding box
            of given EOPatch. If given EOPatch is not provided it will load the entire dataset.
        :return: An EOPatch with an additional vector feature
        """
        eopatch = eopatch or EOPatch()
        bbox = bbox or eopatch.bbox

        if not eopatch.bbox:
            eopatch.bbox = bbox

        vectors = self._load_vector_data(bbox)
        eopatch[self.feature] = self._reproject_and_clip(vectors, bbox)

        return eopatch


class VectorImportTask(_BaseVectorImportTask):
    """A task for importing (Fiona readable) vector data files into an EOPatch"""

    def __init__(
        self,
        feature: FeatureSpec,
        path: str,
        reproject: bool = True,
        clip: bool = False,
        filesystem: Optional[FS] = None,
        config: Optional[SHConfig] = None,
        **kwargs: Any,
    ):
        """
        :param feature: A vector feature into which to import data
        :param path: A path to a dataset containing vector data. It can be either a local path or a path to s3 bucket.
            If `filesystem` parameter is given the path should be relative to the filesystem, otherwise it should
            be an absolute path.
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :param filesystem: A filesystem object. If not given it will be created from the path and config credentials.
        :param config: A configuration object with AWS credentials (if not provided, ~/.aws/credentials will be used)
        :param kwargs: Additional args that will be passed to `fiona.open` or `geopandas.read` calls (e.g. layer name)
        """
        if filesystem is None:
            filesystem, path = get_base_filesystem_and_path(path, config=config)
        self.path = path
        self.full_path = get_full_path(filesystem, path)
        self._pickled_filesystem = pickle_fs(filesystem)

        self.fiona_kwargs = kwargs
        self._aws_session = None
        self._dataset_crs: Optional[CRS] = None

        super().__init__(feature=feature, reproject=reproject, clip=clip, config=config)

    @property
    def aws_session(self) -> AWSSession:
        """Because the session object cannot be pickled this provides the session lazily (i.e. the first time it is
        needed)

        :return: A session for AWS services
        """
        if self._aws_session is None:
            filesystem = unpickle_fs(self._pickled_filesystem)
            if not isinstance(filesystem, S3FS):
                raise ValueError(f"AWS session can only be obtained for S3 filesystem but found {filesystem}")

            boto_session = boto3.session.Session(
                aws_access_key_id=filesystem.aws_access_key_id,
                aws_secret_access_key=filesystem.aws_secret_access_key,
                aws_session_token=filesystem.aws_session_token,
                region_name=filesystem.region,
            )
            self._aws_session = AWSSession(boto_session)

        return self._aws_session

    @property
    def dataset_crs(self) -> Optional[CRS]:
        """Provides a CRS of dataset, it loads it lazily (i.e. the first time it is needed)

        :return: Dataset's CRS
        """
        if self._dataset_crs is None:
            if self.full_path.startswith("s3://"):
                with fiona.Env(session=self.aws_session):
                    self._read_crs()
            else:
                self._read_crs()

        return self._dataset_crs

    def _read_crs(self) -> None:
        """Reads information about CRS from a dataset"""
        with fiona.open(self.full_path, **self.fiona_kwargs) as features:
            self._dataset_crs = CRS(features.crs)

    def _load_vector_data(self, bbox: Optional[BBox]) -> gpd.GeoDataFrame:
        """Loads vector data either from S3 or local path"""
        bbox_bounds = bbox.transform_bounds(self.dataset_crs).geometry.bounds if bbox and self.dataset_crs else None

        if self.full_path.startswith("s3://"):
            with fiona.Env(session=self.aws_session), fiona.open(self.full_path, **self.fiona_kwargs) as features:
                feature_iter = features if bbox_bounds is None else features.filter(bbox=bbox_bounds)

                return gpd.GeoDataFrame.from_features(
                    feature_iter,
                    columns=list(features.schema["properties"]) + ["geometry"],
                    crs=self.dataset_crs.pyproj_crs() if self.dataset_crs else None,
                )

        return gpd.read_file(self.full_path, bbox=bbox_bounds, **self.fiona_kwargs)


class GeopediaVectorImportTask(_BaseVectorImportTask):
    """A task for importing `Geopedia <https://geopedia.world>`__ features into EOPatch vector features"""

    def __init__(
        self,
        feature: FeatureSpec,
        geopedia_table: Union[str, int],
        reproject: bool = True,
        clip: bool = False,
        **kwargs: Any,
    ):
        """
        :param feature: A vector feature into which to import data
        :param geopedia_table: A Geopedia table from which to retrieve features
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :param kwargs: Additional args that will be passed to `GeopediaFeatureIterator`
        """
        self.geopedia_table = geopedia_table
        self.geopedia_kwargs = kwargs
        self.dataset_crs: Optional[CRS] = None
        super().__init__(feature=feature, reproject=reproject, clip=clip)

    def _load_vector_data(self, bbox: Optional[BBox]) -> gpd.GeoDataFrame:
        """Loads vector data from geopedia table"""
        prepared_bbox = bbox.transform_bounds(CRS.POP_WEB) if bbox else None

        geopedia_iterator = GeopediaFeatureIterator(
            layer=self.geopedia_table,
            bbox=prepared_bbox,
            offset=0,
            gpd_session=None,
            config=self.config,
            **self.geopedia_kwargs,
        )
        geopedia_features = list(geopedia_iterator)

        geometry = geopedia_features[0].get("geometry")
        if not geometry:
            raise ValueError(f'Geopedia table "{self.geopedia_table}" does not contain geometries!')

        self.dataset_crs = CRS(geometry["crs"]["properties"]["name"])  # always WGS84
        return gpd.GeoDataFrame.from_features(geopedia_features, crs=self.dataset_crs.pyproj_crs())



"""
Module with tasks that integrate with GeoDB

To use tasks from this module you have to install dependencies defined in `requirements-geodb.txt`.

Credits:
Copyright (c) 2021-2022 Matej Aleksandrov, Matej Batič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

from sentinelhub import CRS

from ..geometry_io import _BaseVectorImportTask


class GeoDBVectorImportTask(_BaseVectorImportTask):
    """A task for importing vector data from `geoDB <https://eurodatacube.com/marketplace/services/edc_geodb>`__
    into EOPatch
    """

    def __init__(self, feature, geodb_client, geodb_collection, geodb_db, reproject=True, clip=False, **kwargs):
        """
        :param feature: A vector feature into which to import data
        :type feature: (FeatureType, str)
        :param geodb_client: an instance of GeoDBClient
        :type geodb_client: xcube_geodb.core.geodb.GeoDBClient
        :param geodb_collection: The name of the collection to be queried
        :type geodb_collection: str
        :param geodb_db: The name of the database the collection resides in [current database]
        :type geodb_db: str
        :param reproject: Should the geometries be transformed to coordinate reference system of the requested bbox?
        :type reproject: bool, default = True
        :param clip: Should the geometries be clipped to the requested bbox, or should be geometries kept as they are?
        :type clip: bool, default = False
        :param kwargs: Additional args that will be passed to `geodb_client.get_collection_by_bbox` call
            (e.g. where="id>-1", operator="and")
        """
        self.geodb_client = geodb_client
        self.geodb_db = geodb_db
        self.geodb_collection = geodb_collection
        self.geodb_kwargs = kwargs
        self._dataset_crs = None

        super().__init__(feature=feature, reproject=reproject, clip=clip)

    @property
    def dataset_crs(self):
        """Provides a "crs" of dataset, loads it lazily (i.e. the first time it is needed)

        :return: Dataset's CRS
        :rtype: CRS
        """
        if self._dataset_crs is None:
            srid = self.geodb_client.get_collection_srid(collection=self.geodb_collection, database=self.geodb_db)
            self._dataset_crs = CRS(f"epsg:{srid}")

        return self._dataset_crs

    def _load_vector_data(self, bbox):
        """Loads vector data from geoDB table"""
        prepared_bbox = bbox.transform_bounds(self.dataset_crs).geometry.bounds if bbox else None

        if "comparison_mode" not in self.geodb_kwargs:
            self.geodb_kwargs["comparison_mode"] = "intersects"

        return self.geodb_client.get_collection_by_bbox(
            collection=self.geodb_collection,
            database=self.geodb_db,
            bbox=prepared_bbox,
            bbox_crs=self.dataset_crs.epsg,
            **self.geodb_kwargs,
        )



""" Input tasks that collect data from `Sentinel-Hub Process API
<https://docs.sentinel-hub.com/api/latest/api/process/>`__

Credits:
Copyright (c) 2019-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2019-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2019-2021 Beno Šircelj

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
from __future__ import annotations

import datetime as dt
import logging
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np

from sentinelhub import (
    BBox,
    DataCollection,
    Geometry,
    MimeType,
    MosaickingOrder,
    ResamplingType,
    SentinelHubCatalog,
    SentinelHubDownloadClient,
    SentinelHubRequest,
    SentinelHubSession,
    SHConfig,
    bbox_to_dimensions,
    filter_times,
    parse_time_interval,
)
from sentinelhub.types import RawTimeIntervalType

from eolearn.core import EOPatch, EOTask, FeatureType, FeatureTypeSet
from eolearn.core.types import FeatureRenameSpec, FeatureSpec, FeaturesSpecification, Literal

LOGGER = logging.getLogger(__name__)


class SentinelHubInputBaseTask(EOTask):
    """Base class for Processing API input tasks"""

    def __init__(
        self,
        data_collection: DataCollection,
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Union[float, Tuple[float, float]]] = None,
        cache_folder: Optional[str] = None,
        config: Optional[SHConfig] = None,
        max_threads: Optional[int] = None,
        upsampling: Optional[ResamplingType] = None,
        downsampling: Optional[ResamplingType] = None,
        session_loader: Optional[Callable[[], SentinelHubSession]] = None,
    ):
        """
        :param data_collection: A collection of requested satellite data.
        :param size: Number of pixels in x and y dimension.
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :param config: An instance of SHConfig defining the service
        :param max_threads: Maximum threads to be used when downloading data.
        :param upsampling: A type of upsampling to apply on data
        :param downsampling: A type of downsampling to apply on data
        :param session_loader: A callable that returns a valid SentinelHubSession, used for session sharing.
            Creates a new session if set to `None`, which should be avoided in large scale parallelization.
        """
        if (size is None) == (resolution is None):
            raise ValueError("Exactly one of the parameters 'size' and 'resolution' should be given.")

        self.size = size
        self.resolution = resolution
        self.config = config or SHConfig()
        self.max_threads = max_threads
        self.data_collection: DataCollection = DataCollection(data_collection)
        self.cache_folder = cache_folder
        self.session_loader = session_loader
        self.upsampling = upsampling
        self.downsampling = downsampling

    def execute(
        self,
        eopatch: Optional[EOPatch] = None,
        bbox: Optional[BBox] = None,
        time_interval: Optional[RawTimeIntervalType] = None,  # should be kept at this to prevent code-breaks
        geometry: Optional[Geometry] = None,
    ) -> EOPatch:
        """Main execute method for the Process API tasks.
        The `geometry` is used only in conjunction with the `bbox` and does not act as a replacement."""

        eopatch = eopatch or EOPatch()

        eopatch.bbox = self._extract_bbox(bbox, eopatch)
        size_x, size_y = self._get_size(eopatch)

        if time_interval:
            time_interval = parse_time_interval(time_interval)
            timestamp = self._get_timestamp(time_interval, eopatch.bbox)
            timestamp = [time_point.replace(tzinfo=None) for time_point in timestamp]
        elif self.data_collection.is_timeless:
            timestamp = None  # should this be [] to match next branch in case of a fresh eopatch?
        else:
            timestamp = eopatch.timestamp

        if timestamp is not None:
            if not eopatch.timestamp:
                eopatch.timestamp = timestamp
            elif timestamp != eopatch.timestamp:
                raise ValueError("Trying to write data to an existing EOPatch with a different timestamp.")

        sh_requests = self._build_requests(eopatch.bbox, size_x, size_y, timestamp, time_interval, geometry)
        requests = [request.download_list[0] for request in sh_requests]

        LOGGER.debug("Downloading %d requests of type %s", len(requests), str(self.data_collection))
        session = None if self.session_loader is None else self.session_loader()
        client = SentinelHubDownloadClient(config=self.config, session=session)
        responses = client.download(requests, max_threads=self.max_threads)
        LOGGER.debug("Downloads complete")

        temporal_dim = 1 if timestamp is None else len(timestamp)
        shape = temporal_dim, size_y, size_x
        self._extract_data(eopatch, responses, shape)

        return eopatch

    def _get_size(self, eopatch: EOPatch) -> Tuple[int, int]:
        """Get the size (width, height) for the request either from inputs, or from the (existing) eopatch"""
        if self.size is not None:
            return self.size

        if self.resolution is not None:
            return bbox_to_dimensions(eopatch.bbox, self.resolution)

        raise ValueError("Size or resolution for the requests should be provided!")

    @staticmethod
    def _extract_bbox(bbox: Optional[BBox], eopatch: EOPatch) -> BBox:
        if eopatch.bbox is None:
            if bbox is None:
                raise ValueError("Either the eopatch or the task must provide valid bbox.")
            return bbox

        if bbox is None or eopatch.bbox == bbox:
            return eopatch.bbox
        raise ValueError("Either the eopatch or the task must provide bbox, or they must be the same.")

    def _extract_data(self, eopatch, images, shape):
        """Extract data from the received images and assign them to eopatch features"""
        raise NotImplementedError("The _extract_data method should be implemented by the subclass.")

    def _build_requests(
        self,
        bbox: Optional[BBox],
        size_x: int,
        size_y: int,
        timestamp: Optional[List[dt.datetime]],
        time_interval: Optional[RawTimeIntervalType],
        geometry: Optional[Geometry],
    ) -> List[SentinelHubRequest]:
        """Build requests"""
        raise NotImplementedError("The _build_requests method should be implemented by the subclass.")

    def _get_timestamp(self, time_interval: Optional[RawTimeIntervalType], bbox: BBox) -> List[dt.datetime]:
        """Get the timestamp array needed as a parameter for downloading the images"""
        raise NotImplementedError("The _get_timestamp method should be implemented by the subclass.")


class SentinelHubEvalscriptTask(SentinelHubInputBaseTask):
    """Process API task to download data using evalscript"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        features: FeaturesSpecification,
        evalscript: str,
        data_collection: DataCollection,
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Union[float, Tuple[float, float]]] = None,
        maxcc: Optional[float] = None,
        time_difference: Optional[dt.timedelta] = None,
        mosaicking_order: Optional[Union[str, MosaickingOrder]] = None,
        cache_folder: Optional[str] = None,
        config: Optional[SHConfig] = None,
        max_threads: Optional[int] = None,
        upsampling: Optional[ResamplingType] = None,
        downsampling: Optional[ResamplingType] = None,
        aux_request_args: Optional[dict] = None,
        session_loader: Optional[Callable[[], SentinelHubSession]] = None,
        timestamp_filter: Callable[[List[dt.datetime], dt.timedelta], List[dt.datetime]] = filter_times,
    ):
        """
        :param features: Features to construct from the evalscript.
        :param evalscript: Evalscript for the request. Beware that all outputs from SentinelHub services should be named
            and should have the same name as corresponding feature
        :param data_collection: Source of requested satellite data.
        :param size: Number of pixels in x and y dimension.
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :param maxcc: Maximum cloud coverage, a float in interval [0, 1]
        :param time_difference: Minimum allowed time difference, used when filtering dates. Also used by the service
            for mosaicking, timestamps might be misleading for large values.
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :param config: An instance of SHConfig defining the service
        :param max_threads: Maximum threads to be used when downloading data.
        :param upsampling: A type of upsampling to apply on data
        :param downsampling: A type of downsampling to apply on data
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :param aux_request_args: a dictionary with auxiliary information for the input_data part of the SH request
        :param session_loader: A callable that returns a valid SentinelHubSession, used for session sharing.
            Creates a new session if set to `None`, which should be avoided in large scale parallelization.
        :param timestamp_filter: A function that performs the final filtering of timestamps, usually to remove multiple
            occurrences within the time_difference window. Check `get_available_timestamps` for more info.
        """
        super().__init__(
            data_collection=data_collection,
            size=size,
            resolution=resolution,
            cache_folder=cache_folder,
            config=config,
            max_threads=max_threads,
            upsampling=upsampling,
            downsampling=downsampling,
            session_loader=session_loader,
        )

        self.features = self._parse_and_validate_features(features)
        self.responses = self._create_response_objects()
        self.evalscript = evalscript

        if maxcc and isinstance(maxcc, (int, float)) and (maxcc < 0 or maxcc > 1):
            raise ValueError("maxcc should be a float on an interval [0, 1]")

        self.maxcc = maxcc
        self.time_difference = time_difference or dt.timedelta(seconds=1)
        self.timestamp_filter = timestamp_filter
        self.mosaicking_order = None if mosaicking_order is None else MosaickingOrder(mosaicking_order)
        self.aux_request_args = aux_request_args

    def _parse_and_validate_features(self, features: FeaturesSpecification) -> List[FeatureRenameSpec]:
        allowed_features = FeatureTypeSet.RASTER_TYPES.union({FeatureType.META_INFO})
        _features = self.parse_renamed_features(features, allowed_feature_types=allowed_features)

        ftr_data_types = {ft for ft, _, _ in _features if not ft.is_meta()}
        if all(ft.is_timeless() for ft in ftr_data_types) or all(ft.is_temporal() for ft in ftr_data_types):
            return _features

        raise ValueError("Cannot mix time dependent and timeless requests!")

    def _create_response_objects(self):
        """Construct SentinelHubRequest output_responses from features"""
        responses = []
        for feat_type, feat_name, _ in self.features:
            if feat_type.is_raster():
                responses.append(SentinelHubRequest.output_response(feat_name, MimeType.TIFF))
            elif feat_type.is_meta():
                responses.append(SentinelHubRequest.output_response("userdata", MimeType.JSON))
            else:
                # should not happen as features have already been validated
                raise ValueError(f"{feat_type} not supported!")

        return responses

    def _get_timestamp(self, time_interval: Optional[RawTimeIntervalType], bbox: BBox) -> List[dt.datetime]:
        """Get the timestamp array needed as a parameter for downloading the images"""
        if any(feat_type.is_timeless() for feat_type, _, _ in self.features if feat_type.is_raster()):
            return []

        return get_available_timestamps(
            bbox=bbox,
            time_interval=time_interval,
            timestamp_filter=self.timestamp_filter,
            data_collection=self.data_collection,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            config=self.config,
        )

    def _build_requests(
        self,
        bbox: Optional[BBox],
        size_x: int,
        size_y: int,
        timestamp: Optional[List[dt.datetime]],
        time_interval: Optional[RawTimeIntervalType],
        geometry: Optional[Geometry],
    ):
        """Defines request timestamps and builds requests. In case `timestamp` is either `None` or an empty list it
        still has to create at least one request in order to obtain back number of bands of responses."""
        if timestamp:
            dates = [(date - self.time_difference, date + self.time_difference) for date in timestamp]
        elif timestamp is None:
            dates = [None]
        else:
            dates = [parse_time_interval(time_interval, allow_undefined=True)]

        return [self._create_sh_request(date, bbox, size_x, size_y, geometry) for date in dates]

    def _create_sh_request(self, time_interval, bbox, size_x, size_y, geometry):
        """Create an instance of SentinelHubRequest"""
        return SentinelHubRequest(
            evalscript=self.evalscript,
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=self.data_collection,
                    mosaicking_order=self.mosaicking_order,
                    time_interval=time_interval,
                    maxcc=self.maxcc,
                    upsampling=self.upsampling,
                    downsampling=self.downsampling,
                    other_args=self.aux_request_args,
                )
            ],
            responses=self.responses,
            bbox=bbox,
            geometry=geometry,
            size=(size_x, size_y),
            data_folder=self.cache_folder,
            config=self.config,
        )

    def _extract_data(self, eopatch, data_responses, shape):
        """Extract data from the received images and assign them to eopatch features"""
        # pylint: disable=arguments-renamed
        if len(self.features) == 1:
            ftype, fname, _ = self.features[0]
            extension = "json" if ftype.is_meta() else "tif"
            data_responses = [{f"{fname}.{extension}": data} for data in data_responses]

        for ftype, fname, new_fname in self.features:
            if ftype.is_meta():
                data = [data["userdata.json"] for data in data_responses]

            elif ftype.is_temporal():
                data = np.asarray([data[f"{fname}.tif"] for data in data_responses])
                data = data[..., np.newaxis] if data.ndim == 3 else data
                time_dim = shape[0]
                data = data[:time_dim] if time_dim != data.shape[0] else data

            else:
                data = np.asarray(data_responses[0][f"{fname}.tif"])[..., np.newaxis]

            eopatch[ftype][new_fname] = data

        return eopatch


class SentinelHubInputTask(SentinelHubInputBaseTask):
    """Process API input task that loads 16bit integer data and converts it to a 32bit float feature."""

    # pylint: disable=too-many-arguments
    DTYPE_TO_SAMPLE_TYPE = {
        bool: "SampleType.UINT8",
        np.uint8: "SampleType.UINT8",
        np.uint16: "SampleType.UINT16",
        np.float32: "SampleType.FLOAT32",
    }

    # pylint: disable=too-many-locals
    def __init__(
        self,
        data_collection: DataCollection,
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Union[float, Tuple[float, float]]] = None,
        bands_feature: Optional[Tuple[FeatureType, str]] = None,
        bands: Optional[List[str]] = None,
        additional_data: Optional[List[Tuple[FeatureType, str]]] = None,
        evalscript: Optional[str] = None,
        maxcc: Optional[float] = None,
        time_difference: Optional[dt.timedelta] = None,
        cache_folder: Optional[str] = None,
        config: Optional[SHConfig] = None,
        max_threads: Optional[int] = None,
        bands_dtype: Union[None, np.dtype, type] = None,
        single_scene: bool = False,
        mosaicking_order: Optional[Union[str, MosaickingOrder]] = None,
        upsampling: Optional[ResamplingType] = None,
        downsampling: Optional[ResamplingType] = None,
        aux_request_args: Optional[dict] = None,
        session_loader: Optional[Callable[[], SentinelHubSession]] = None,
        timestamp_filter: Callable[[List[dt.datetime], dt.timedelta], List[dt.datetime]] = filter_times,
    ):
        """
        :param data_collection: Source of requested satellite data.
        :param size: Number of pixels in x and y dimension.
        :param resolution: Resolution in meters, passed as a single number or a tuple of two numbers -
            resolution in horizontal and resolution in vertical direction.
        :param bands_feature: A target feature into which to save the downloaded images.
        :param bands: An array of band names. If not specified it will download all bands specified for a given data
            collection.
        :param additional_data: A list of additional data to be downloaded, such as SCL, SNW, dataMask, etc.
        :param evalscript: An optional parameter to override an evalscript that is generated by default
        :param maxcc: Maximum cloud coverage.
        :param time_difference: Minimum allowed time difference, used when filtering dates. Also used by the service
            for mosaicking, timestamps might be misleading for large values.
        :param cache_folder: Path to cache_folder. If set to None (default) requests will not be cached.
        :param config: An instance of SHConfig defining the service
        :param max_threads: Maximum threads to be used when downloading data.
        :param bands_dtype: output type of the bands array, if set to None the default is used
        :param single_scene: If true, the service will compute a single image for the given time interval using
            mosaicking.
        :param mosaicking_order: Mosaicking order, which has to be either 'mostRecent', 'leastRecent' or 'leastCC'.
        :param upsampling: A type of upsampling to apply on data
        :param downsampling: A type of downsampling to apply on data
        :param aux_request_args: a dictionary with auxiliary information for the input_data part of the SH request
        :param session_loader: A callable that returns a valid SentinelHubSession, used for session sharing.
            Creates a new session if set to `None`, which should be avoided in large scale parallelization.
        :param timestamp_filter: A callable that performs the final filtering of timestamps, usually to remove multiple
            occurrences within the time_difference window. Check `get_available_timestamps` for more info.
        """
        super().__init__(
            data_collection=data_collection,
            size=size,
            resolution=resolution,
            cache_folder=cache_folder,
            config=config,
            max_threads=max_threads,
            upsampling=upsampling,
            downsampling=downsampling,
            session_loader=session_loader,
        )
        self.evalscript = evalscript
        self.maxcc = maxcc
        self.time_difference = time_difference or dt.timedelta(seconds=1)
        self.timestamp_filter = timestamp_filter
        self.single_scene = single_scene
        self.bands_dtype = bands_dtype
        self.mosaicking_order = None if mosaicking_order is None else MosaickingOrder(mosaicking_order)
        self.aux_request_args = aux_request_args

        self.bands_feature = None
        self.requested_bands = []
        if bands_feature:
            self.bands_feature = self.parse_feature(bands_feature, allowed_feature_types=[FeatureType.DATA])
            if bands is not None:
                self.requested_bands = self._parse_requested_bands(bands, self.data_collection.bands)
            else:
                self.requested_bands = list(self.data_collection.bands)

        self.requested_additional_bands = []
        if additional_data is not None:
            additional_data = self.parse_renamed_features(additional_data)
            additional_bands = [band for _, band, _ in additional_data]
            parsed_bands = self._parse_requested_bands(additional_bands, self.data_collection.metabands)
            self.requested_additional_bands = parsed_bands

        self.additional_data = additional_data

    def _parse_requested_bands(self, bands, available_bands):
        """Checks that all requested bands are available and returns the band information for further processing"""
        requested_bands = []
        band_info_dict = {band_info.name: band_info for band_info in available_bands}
        for band_name in bands:
            if band_name in band_info_dict:
                requested_bands.append(band_info_dict[band_name])
            else:
                raise ValueError(
                    f"Data collection {self.data_collection} does not have specifications for {band_name}."
                    f"Available bands are {[band.name for band in self.data_collection.bands]} and meta-bands"
                    f"{[band.name for band in self.data_collection.metabands]}"
                )
        return requested_bands

    def generate_evalscript(self) -> str:
        """Generate the evalscript to be passed with the request, based on chosen bands"""
        evalscript = """
            //VERSION=3

            function setup() {{
                return {{
                    input: [{{
                        bands: [{bands}],
                        units: [{units}]
                    }}],
                    output: [
                        {outputs}
                    ]
                }}
            }}

            function evaluatePixel(sample) {{
                return {{ {samples} }}
            }}
        """

        bands, units, outputs, samples = [], [], [], []
        for band in self.requested_bands + self.requested_additional_bands:
            unit_choice = 0  # use default units
            if band in self.requested_bands and self.bands_dtype is not None:
                if self.bands_dtype not in band.output_types:
                    raise ValueError(
                        f"Band {band.name} only supports output types {band.output_types} but `bands_dtype` is set to "
                        f"{self.bands_dtype}. To use default types set `bands_dtype` to None."
                    )
                unit_choice = band.output_types.index(self.bands_dtype)

            sample_type = SentinelHubInputTask.DTYPE_TO_SAMPLE_TYPE[band.output_types[unit_choice]]

            bands.append(f'"{band.name}"')
            units.append(f'"{band.units[unit_choice].value}"')
            samples.append(f"{band.name}: [sample.{band.name}]")
            outputs.append(f'{{ id: "{band.name}", bands: 1, sampleType: {sample_type} }}')

        evalscript = evalscript.format(
            bands=", ".join(bands), units=", ".join(units), outputs=", ".join(outputs), samples=", ".join(samples)
        )

        return evalscript

    def _get_timestamp(self, time_interval: Optional[RawTimeIntervalType], bbox: BBox) -> List[dt.datetime]:
        """Get the timestamp array needed as a parameter for downloading the images"""
        if self.single_scene:
            return [time_interval[0]]

        return get_available_timestamps(
            bbox=bbox,
            time_interval=time_interval,
            timestamp_filter=self.timestamp_filter,
            data_collection=self.data_collection,
            maxcc=self.maxcc,
            time_difference=self.time_difference,
            config=self.config,
        )

    def _build_requests(
        self,
        bbox: Optional[BBox],
        size_x: int,
        size_y: int,
        timestamp: Optional[List[dt.datetime]],
        time_interval: Optional[RawTimeIntervalType],
        geometry: Optional[Geometry],
    ) -> List[SentinelHubRequest]:
        """Build requests"""
        if timestamp is None:
            intervals: List[Optional[RawTimeIntervalType]] = [None]
        elif self.single_scene:
            intervals = [parse_time_interval(time_interval)]
        else:
            intervals = [(date - self.time_difference, date + self.time_difference) for date in timestamp]

        return [self._create_sh_request(time_interval, bbox, size_x, size_y, geometry) for time_interval in intervals]

    def _create_sh_request(self, time_interval, bbox, size_x, size_y, geometry):
        """Create an instance of SentinelHubRequest"""
        responses = [
            SentinelHubRequest.output_response(band.name, MimeType.TIFF)
            for band in self.requested_bands + self.requested_additional_bands
        ]

        return SentinelHubRequest(
            evalscript=self.evalscript or self.generate_evalscript(),
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=self.data_collection,
                    time_interval=time_interval,
                    mosaicking_order=self.mosaicking_order,
                    maxcc=self.maxcc,
                    upsampling=self.upsampling,
                    downsampling=self.downsampling,
                    other_args=self.aux_request_args,
                )
            ],
            responses=responses,
            bbox=bbox,
            geometry=geometry,
            size=(size_x, size_y),
            data_folder=self.cache_folder,
            config=self.config,
        )

    def _extract_data(self, eopatch, images, shape):
        """Extract data from the received images and assign them to eopatch features"""
        if len(self.requested_bands) + len(self.requested_additional_bands) == 1:
            # if only one band is requested the response is not a tar so we reshape it
            only_band = (self.requested_bands + self.requested_additional_bands)[0]
            images = [{only_band.name + ".tif": image} for image in images]

        if self.additional_data:
            self._extract_additional_features(eopatch, images, shape)

        if self.bands_feature:
            self._extract_bands_feature(eopatch, images, shape)

        return eopatch

    def _extract_additional_features(self, eopatch, images, shape):
        """Extracts additional features from response into an EOPatch"""
        for (ftype, _, new_name), band_info in zip(self.additional_data, self.requested_additional_bands):
            tiffs = [tar[band_info.name + ".tif"] for tar in images]
            eopatch[ftype, new_name] = self._extract_array(tiffs, 0, shape, band_info.output_types[0])

    def _extract_bands_feature(self, eopatch, images, shape):
        """Extract the bands feature arrays and concatenate them along the last axis"""
        processed_bands = []
        for band_info in self.requested_bands:
            tiffs = [tar[band_info.name + ".tif"] for tar in images]
            dtype = self.bands_dtype or band_info.output_types[0]
            processed_bands.append(self._extract_array(tiffs, 0, shape, dtype))

        eopatch[self.bands_feature] = np.concatenate(processed_bands, axis=-1)

    @staticmethod
    def _extract_array(tiffs, idx, shape, dtype):
        """Extract a numpy array from the received tiffs"""
        feature_arrays = (np.atleast_3d(img)[..., idx] for img in tiffs)
        return np.asarray(list(feature_arrays), dtype=dtype).reshape(*shape, 1)


class SentinelHubDemTask(SentinelHubEvalscriptTask):
    """
    Adds DEM data (one of the `collections <https://docs.sentinel-hub.com/api/latest/data/dem/#deminstance>`__) to
        DATA_TIMELESS EOPatch feature.
    """

    def __init__(
        self,
        feature: Union[None, str, FeatureSpec] = None,
        data_collection: DataCollection = DataCollection.DEM,
        **kwargs: Any,
    ):
        if feature is None:
            feature = (FeatureType.DATA_TIMELESS, "dem")
        elif isinstance(feature, str):
            feature = (FeatureType.DATA_TIMELESS, feature)

        feature_type, feature_name = feature
        if feature_type.is_temporal():
            raise ValueError("DEM feature should be timeless!")

        band = data_collection.bands[0]

        evalscript = f"""
            //VERSION=3

            function setup() {{
                return {{
                    input: [{{
                        bands: ["{band.name}"],
                        units: ["{band.units[0].value}"]
                    }}],
                    output: {{
                        id: "{feature_name}",
                        bands: 1,
                        sampleType: SampleType.UINT16
                    }}
                }}
            }}

            function evaluatePixel(sample) {{
                return {{ {feature_name}: [sample.{band.name}] }}
            }}
        """

        super().__init__(evalscript=evalscript, features=[feature], data_collection=data_collection, **kwargs)


class SentinelHubSen2corTask(SentinelHubInputTask):
    """
    Adds SCL (scene classification), CLD (cloud probability) or SNW (snow probability) (or their combination)
    Sen2Cor classification results to EOPatch's MASK or DATA feature. The feature is added to MASK (SCL) or
    DATA (CLD, SNW) feature types of EOPatch. The feature names are set to be SCL, CLD or SNW.

    Sen2Cor's scene classification (SCL) contains 11 classes with the following values and meanings:
       * 1 - SC_SATURATED_DEFECTIVE
       * 2 - SC_DARK_FEATURE_SHADOW
       * 3 - SC_CLOUD_SHADOW
       * 4 - VEGETATION
       * 5 - NOT-VEGETATED
       * 6 - WATER
       * 7 - SC_CLOUD_LOW_PROBA / UNCLASSIFIED
       * 8 - SC_CLOUD_MEDIUM_PROBA
       * 9 - CLOUD_HIGH_PROBABILITY
       * 10 - THIN_CIRRUS
       * 11 - SNOW
    """

    def __init__(
        self,
        sen2cor_classification: Union[Literal["SCL", "CLD", "SNW"], List[Literal["SCL", "CLD", "SNW"]]],
        data_collection: DataCollection = DataCollection.SENTINEL2_L2A,
        **kwargs: Any,
    ):
        """
        :param sen2cor_classification: "SCL" (scene classification), "CLD" (cloud probability) or "SNW"
            (snow probability) masks to be retrieved. Also, a list of their combination (e.g. ["SCL","CLD"])
        :param kwargs: Additional arguments that will be passed to the `SentinelHubInputTask`
        """
        # definition of possible types and target features
        classification_types = {"SCL": FeatureType.MASK, "CLD": FeatureType.DATA, "SNW": FeatureType.DATA}

        if isinstance(sen2cor_classification, str):
            sen2cor_classification = [sen2cor_classification]

        for s2c in sen2cor_classification:
            if s2c not in classification_types:
                raise ValueError(
                    f"Unsupported Sen2Cor classification type: {s2c}. Possible types are: {classification_types}!"
                )

        if data_collection != DataCollection.SENTINEL2_L2A:
            raise ValueError("Sen2Cor classification layers are only available on Sentinel-2 L2A data.")

        features: List[Tuple[FeatureType, str]] = [(classification_types[s2c], s2c) for s2c in sen2cor_classification]
        super().__init__(additional_data=features, data_collection=data_collection, **kwargs)


def get_available_timestamps(
    bbox: BBox,
    data_collection: DataCollection,
    *,
    time_interval: Optional[RawTimeIntervalType] = None,
    time_difference: dt.timedelta = dt.timedelta(seconds=-1),  # noqa: B008
    timestamp_filter: Callable[[List[dt.datetime], dt.timedelta], List[dt.datetime]] = filter_times,
    maxcc: Optional[float] = None,
    config: Optional[SHConfig] = None,
) -> List[dt.datetime]:
    """Helper function to search for all available timestamps, based on query parameters.

    :param bbox: A bounding box of the search area.
    :param data_collection: A data collection for which to find available timestamps.
    :param time_interval: A time interval from which to provide the timestamps.
    :param time_difference: Minimum allowed time difference, used when filtering dates.
    :param timestamp_filter: A function that performs the final filtering of timestamps, usually to remove multiple
        occurrences within the time_difference window. The filtration is performed after all suitable timestamps for
        the given region are obtained (with maxcc filtering already done by SH). By default only keeps the oldest
        timestamp when multiple occur within `time_difference`.
    :param maxcc: Maximum cloud coverage filter from interval [0, 1], default is None.
    :param config: A configuration object.
    :return: A list of timestamps of available observations.
    """
    query_filter = None
    if maxcc is not None and data_collection.has_cloud_coverage:
        if isinstance(maxcc, (int, float)) and (maxcc < 0 or maxcc > 1):
            raise ValueError('Maximum cloud coverage "maxcc" parameter should be a float on an interval [0, 1]')
        query_filter = f"eo:cloud_cover < {int(maxcc * 100)}"

    fields = {"include": ["properties.datetime"], "exclude": []}

    if data_collection.service_url:
        config = config.copy() if config else SHConfig()
        config.sh_base_url = data_collection.service_url

    catalog = SentinelHubCatalog(config=config)
    search_iterator = catalog.search(
        collection=data_collection, bbox=bbox, time=time_interval, filter=query_filter, fields=fields
    )

    all_timestamps = search_iterator.get_timestamps()
    return timestamp_filter(all_timestamps, time_difference)
